﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using DocumentFormat.OpenXml.Packaging;
using OpenXmlPowerTools;

namespace XlsxCellDimensions
{
    class Program
    {
        static void GetNormalFontInfo(SpreadsheetDocument sDoc, out string familyName, out string fontSize)
        {
            familyName = null;
            fontSize = null;

            WorkbookStylesPart stylesPart = sDoc
                .WorkbookPart
                .WorkbookStylesPart;
            XDocument stylesXDoc = stylesPart.GetXDocument();
            XElement normalCellStyle = stylesXDoc
                .Root
                .Elements(S.cellStyles)
                .Elements(S.cellStyle)
                .FirstOrDefault(cs => (string)cs.Attribute(SSNoNamespace.name) == "Normal");
            if (normalCellStyle != null)
            {
                string xfId = (string)normalCellStyle.Attribute(SSNoNamespace.xfId);
                XElement xf = stylesXDoc
                    .Root
                    .Elements(S.cellXfs)
                    .Elements(S.xf)
                    .FirstOrDefault(x => (string)x.Attribute(SSNoNamespace.xfId) == xfId);
                if (xf != null)
                {
                    int? fontId = (int?)xf.Attribute(SSNoNamespace.fontId);
                    if (fontId != null)
                    {
                        XElement font = stylesXDoc
                            .Root
                            .Elements(S.fonts)
                            .Elements(S.font)
                            .Skip((int)fontId)
                            .FirstOrDefault();
                        if (font != null)
                        {
                            familyName = (string)font
                                .Elements(S.name)
                                .Attributes(SSNoNamespace.val)
                                .FirstOrDefault();
                            fontSize = (string)font
                                .Elements(S.sz)
                                .Attributes(SSNoNamespace.val)
                                .FirstOrDefault();
                            return;
                        }
                    }
                }
            }
        }

        static int[] GetColumnWidthsInPixels(SpreadsheetDocument sDoc, WorksheetPart worksheet,
            int maxColumn)
        {
            string familyName;
            string fontSizeString;
            GetNormalFontInfo(sDoc, out familyName, out fontSizeString);

            float fontSize;
            fontSize = float.Parse(fontSizeString);
            Font normalFont = new Font(familyName, fontSize, GraphicsUnit.Point);

            // Need a Graphics object to measure font width
            float digitWidth = 0;
            Bitmap bm = new Bitmap(10, 10);
            bm.SetResolution(96, 96);

            Graphics g = Graphics.FromImage(bm);
            SizeF sf = g.MeasureString("0", normalFont, Int32.MaxValue,
                StringFormat.GenericTypographic);
            digitWidth = (int)sf.Width;
            g.Dispose();

            XDocument wsXDoc = worksheet.GetXDocument();

            List<float> charWidths = Enumerable.Range(1, maxColumn)
                .Select(c =>
                {
                    XElement colWidthElement = wsXDoc.Root
                        .Elements(S.cols)
                        .Elements(S.col)
                        .FirstOrDefault(z => c >= (int)z.Attribute(SSNoNamespace.min) &&
                            c <= (int)z.Attribute(SSNoNamespace.max));
                    if (colWidthElement != null)
                        return (float)colWidthElement.Attribute(SSNoNamespace.width);
                    else
                    {
                        float? defaultColumnWidth = (float?)wsXDoc
                            .Elements(S.sheetFormatPr)
                            .Attributes(SSNoNamespace.defaultColWidth)
                            .FirstOrDefault();
                        if (defaultColumnWidth != null)
                            return (float)defaultColumnWidth;
                        int unroundedPixelWidth = (int)(8 * digitWidth + 5);
                        int roundUpToMultipleOf64 = (unroundedPixelWidth + (8 - unroundedPixelWidth % 8));
                        float charWidth = roundUpToMultipleOf64 / digitWidth;
                        float charWidth2 = ((int)(charWidth * 256) / (float)(256));
                        return charWidth2;
                    }
                })
                .ToList();
            int[] pixelWidths = charWidths
                .Select(cw =>
                {
                    float widthInPixels = (float)Math.Truncate((double)(((cw * 256) +
                        (float)Math.Truncate((double)(128 / digitWidth))) / 256) * digitWidth);
                    return (int)widthInPixels;
                })
                .ToArray();
            return pixelWidths;
        }

        // 72 vertical points = 96 pixels
        static int[] GetRowHeightsInPixels(SpreadsheetDocument sDoc, WorksheetPart worksheet,
            int maxRow)
        {
            XDocument wsXDoc = worksheet.GetXDocument();
            float? defaultRowHeight = (float?)wsXDoc
                .Root
                .Elements(S.sheetFormatPr)
                .Attributes(SSNoNamespace.defaultRowHeight)
                .FirstOrDefault();
            if (defaultRowHeight == null)
                defaultRowHeight = 15;
            Dictionary<int, float> customRowHeights = wsXDoc
                .Root
                .Elements(S.sheetData)
                .Elements(S.row)
                .Where(r => r.Attribute(SSNoNamespace.customHeight) != null ||
                    r.Attribute(SSNoNamespace.thickBot) != null ||
                    r.Attribute(SSNoNamespace.thickTop) != null)
                .ToDictionary(
                    r => (int)r.Attribute(SSNoNamespace.r),
                    r =>
                    {
                        float? h = (float?)r.Attribute(SSNoNamespace.ht);
                        if (h == null)
                            h = defaultRowHeight;
                        return (float)h;
                    });

            int[] rowHeightsInPixels = Enumerable.Range(1, maxRow)
                .Select(rowNum =>
                {
                    float rh = (float)defaultRowHeight;
                    if (customRowHeights.ContainsKey(rowNum))
                        rh = customRowHeights[rowNum];
                    int x = (int)(rh * 20 / 15);
                    return x;
                })
                .ToArray();
            return rowHeightsInPixels;
        }

        static void Main(string[] args)
        {
            using (SpreadsheetDocument sDoc = 
                SpreadsheetDocument.Open("../../Test.xlsx", true))
            {
                string sheet = "Sheet1";
                XDocument workbookXDoc = sDoc.WorkbookPart.GetXDocument();
                string sheet1Rid = (string)workbookXDoc
                    .Root
                    .Elements(S.sheets)
                    .Elements(S.sheet)
                    .Where(sh => (string)sh.Attribute(SSNoNamespace.name) == sheet)
                    .Attributes(R.id).FirstOrDefault();
                WorksheetPart sheetPart = (WorksheetPart)sDoc
                    .WorkbookPart
                    .GetPartById(sheet1Rid);
                if (sheetPart != null)
                {
                    int[] colWidths = GetColumnWidthsInPixels(sDoc, sheetPart, 10);
                    for (int i = 0; i < colWidths.Length; ++i)
                        Console.WriteLine("Column {0}: {1}", i + 1, colWidths[i]);

                    int[] rowHeights = GetRowHeightsInPixels(sDoc, sheetPart, 10);
                    for (int i = 0; i < rowHeights.Length; ++i)
                        Console.WriteLine("Row {0}: {1}", i + 1, rowHeights[i]);
                }
            }
        }
    }
}
